from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
import warnings as w
from scipy.optimize import root_scalar


class carbon(model_classes.Model):
    """
    Implementation of the Allison et al. (2019) CO2 solubility model. Which type of fit, and
    which composition must be selected when the Model is initialized. The fit may be either
    thermodynamic or power-law. The composition may be chosen from sunset, sfvf, erebus, vesuvius,
    etna, or stromboli. Default is the power-law fit to sunset.
    """

    def __init__(self, model_loc='sunset', model_fit='thermodynamic'):
        """
        Initialize the model.

        Parameters
        ----------
        model_fit     str
            Either 'power' for the power-law fits, or 'thermodynamic' for the
            thermodynamic fits.
        model_loc     str
            One of 'sunset', 'sfvf', 'erebus', 'vesuvius', 'etna', 'stromboli'.
        """

        self.set_volatile_species(['CO2'])
        self.set_fugacity_model(fugacity_models.fugacity_HB_co2())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'temperature', 1200, calibration_checks.crf_EqualTo, 'oC',
                'Allison et al. (2019) carbon', fail_msg=crmsg_Temp,
                pass_msg=calibration_checks.crmsg_EqualTo_pass,
                description_msg=calibration_checks.crmsg_EqualTo_description),
            calibration_checks.CalibrationRange(
                'temperature', [1000, 1400], calibration_checks.crf_Between, 'oC',
                'Allison et al. (2019) carbon', fail_msg=crmsg_Between_Temp,
                pass_msg=calibration_checks.crmsg_EqualTo_pass,
                description_msg=calibration_checks.crmsg_EqualTo_description),
            calibration_checks.CalibrationRange(
                'H2O', 0.5, calibration_checks.crf_LessThan, 'wt%', 'Allison et al. (2019) carbon',
                fail_msg=crmsg_H2O, pass_msg=calibration_checks.crmsg_LessThan_pass)])
        self.set_solubility_dependence(False)
        self.model_loc = model_loc
        self.model_fit = model_fit

    def calculate_dissolved_volatiles(self, pressure, temperature=1200, sample=None, X_fluid=1.0,
                                      **kwargs):
        """
        Calclates the dissolved CO2 concentration using (Eqns) 2-7 or 10-11 from Allison et al.
        (2019).

        Parameters
        ----------
        pressure     float
            Pressure in bars.
        temperature     float
            Temperature in C.
        sample      NoneType or Sample class
            Magma major element composition. Not required for this model, therefore None may be
            passed.
        X_fluid     float
            The mole fraction of CO2 in the fluid. Default is 1.0.

        Returns
        -------
        float
            Dissolved CO2 concentration in wt%.
        """
        # temperature = 1200 #temp in degrees C
        temperature = temperature + 273.15  # translate T from C to K

        if pressure < 0.0:
            raise core.InputError("Pressure must be positive.")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")

        if self.model_fit not in ['power', 'thermodynamic']:
            raise core.InputError("model_fit must be one of 'power', or 'thermodynamic'.")
        if self.model_loc not in ['sunset', 'sfvf', 'erebus', 'vesuvius', 'etna', 'stromboli']:
            raise core.InputError("model_loc must be one of 'sunset', 'sfvf', 'erebus', ",
                                  "'vesuvius', 'etna', or 'stromboli'.")

        if pressure == 0:
            return 0

        if self.model_fit == 'thermodynamic':
            P0 = 1000  # bar
            params = dict({'sunset':    [16.4, -14.67],
                           'sfvf':      [15.02, -14.87],
                           'erebus':    [15.83, -14.65],
                           'vesuvius':  [24.42, -14.04],
                           'etna':      [21.59, -14.28],
                           'stromboli': [14.93, -14.68]})

            DV = params[self.model_loc][0]
            lnK0 = params[self.model_loc][1]

            lnK = lnK0 - (pressure-P0)*DV/(10*8.3141*temperature)
            fCO2 = self.fugacity_model.fugacity(pressure=pressure, temperature=temperature-273.15,
                                                X_fluid=X_fluid, **kwargs)
            Kf = np.exp(lnK)*fCO2
            XCO3 = Kf/(1-Kf)

            FWone = 36.594
            wtCO2 = (44.01*XCO3)/((44.01*XCO3)+(1-XCO3)*FWone)*100
            return wtCO2

        if self.model_fit == 'power':
            params = dict({'stromboli': [1.05, 0.883],
                           'etna':      [2.831, 0.797],
                           'vesuvius':  [4.796, 0.754],
                           'sfvf':      [3.273, 0.74],
                           'sunset':    [4.32, 0.728],
                           'erebus':    [5.145, 0.713]})

            fCO2 = self.fugacity_model.fugacity(pressure=pressure, temperature=temperature-273.15,
                                                X_fluid=X_fluid, **kwargs)

            return params[self.model_loc][0]*fCO2**params[self.model_loc][1]/1e4

    def calculate_equilibrium_fluid_comp(self, pressure, sample, temperature=1200, **kwargs):
        """ Returns 1.0 if a pure CO2 fluid is saturated. Returns 0.0 if a pure CO2 fluid is
        undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        temperature     float
            The temperature of the system in C.
        sample:        Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            1.0 if CO2-fluid saturated, 0.0 otherwise.
        """

        satP = self.calculate_saturation_pressure(temperature=temperature, sample=sample,
                                                  X_fluid=1.0, **kwargs)
        if pressure < satP:
            return 1.0
        else:
            return 0.0

    def calculate_saturation_pressure(self, sample, temperature=1200, X_fluid=1.0, **kwargs):
        """
        Calculates the pressure at which a pure CO2 fluid is saturated, for the given sample
        composition and CO2 concentration. Calls the scipy.root_scalar routine, which makes
        repeated called to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        temperature     float
            The temperature of the system in C.
        sample:        Sample class
            Magma major element composition (including CO2).
        X_fluid     float
            The mole fraction of H2O in the fluid. Default is 1.0.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """

        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('CO2') is False:
            raise core.InputError("sample must contain CO2.")
        if sample.get_composition('CO2') < 0.0:
            raise core.InputError("Dissolved CO2 concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure,
                               args=(temperature, sample, X_fluid, kwargs),
                               x0=1000.0, x1=2000.0).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self, pressure, temperature, sample, X_fluid, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        temperature     float
            The temperature of the system in C.
        sample:     Sample class
            Magma major element composition, including CO2.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved CO2 at the pressure guessed, and the CO2
            concentration passed in the sample variable.
        """
        return (sample.get_composition('CO2') -
                self.calculate_dissolved_volatiles(pressure=pressure, temperature=temperature,
                                                   sample=sample, X_fluid=X_fluid, **kwargs))


# ALLISON COMPOSITIONAL LIMITS -DEFINED AS MIN AND MAX OF CALIBRATION DATASET (-5% AND +5%
# RESPECTIVELY)

def crf_generic(calibval=None, sample=sample_class.Sample({}), bounds={}):
    comp = sample.get_composition(units='wtpt_oxides')
    testresults = []
    for ox in sfvfCompRange:
        testresults.append(comp[ox] >= bounds[ox][0])
        testresults.append(comp[ox] <= bounds[ox][1])
    return all(testresults)


sfvfCompRange = {'SiO2':  [50.0, 55.97],
                 'TiO2':  [1.08, 1.25],
                 'Al2O3': [15.76, 18.15],
                 'FeO':   [7.07, 8.06],
                 'MgO':   [5.89, 7.09],
                 'CaO':   [8.80, 9.91],
                 'Na2O':  [3.05, 3.53],
                 'K2O':   [1.25, 1.49]
                 }


def crf_sfvf(calibval=None, sample=sample_class.Sample({})):
    return crf_generic(calibval, sample, sfvfCompRange)


sunsetCompRange = {'SiO2':  [45.72, 50.62],
                   'TiO2':  [1.75, 1.96],
                   'Al2O3': [15.62, 17.45],
                   'FeO':   [9.13, 10.42],
                   'MgO':   [8.13, 9.19],
                   'CaO':   [9.56, 10.59],
                   'Na2O':  [3.29, 3.64],
                   'K2O':   [0.77, 0.86]
                   }


def crf_sunset(calibval=None, sample={}):
    return crf_generic(calibval, sample, sunsetCompRange)


erebusCompRange = {'SiO2':  [45.96, 51.04],
                   'TiO2':  [2.67, 3.00],
                   'Al2O3': [18.31, 20.63],
                   'FeO':   [7.46, 9.37],
                   'MgO':   [3.03, 3.43],
                   'CaO':   [6.58, 7.42],
                   'Na2O':  [5.8, 6.49],
                   'K2O':   [2.75, 3.13]
                   }


def crf_erebus(calibval=None, sample={}):
    return crf_generic(calibval, sample, erebusCompRange)


vesuviusCompRange = {'SiO2':  [45.65, 53.29],
                     'TiO2':  [0.93, 1.13],
                     'Al2O3': [13.75, 16.28],
                     'FeO':   [4.98, 7.48],
                     'MgO':   [6.41, 7.76],
                     'CaO':   [11.12, 14.16],
                     'Na2O':  [1.74, 2.07],
                     'K2O':   [5.48, 6.35]
                     }


def crf_vesuvius(calibval=None, sample={}):
    return crf_generic(calibval, sample, vesuviusCompRange)


etnaCompRange = {'SiO2':  [46.03, 52.97],
                 'TiO2':  [1.61, 1.89],
                 'Al2O3': [15.87, 18.24],
                 'FeO':   [6.75, 10.21],
                 'MgO':   [5.9, 7.0],
                 'CaO':   [9.38, 11.99],
                 'Na2O':  [3.4, 3.94],
                 'K2O':   [1.69, 2.25]
                 }


def crf_etna(calibval=None, sample={}):
    return crf_generic(calibval, sample, etnaCompRange)


stromboliCompRange = {'SiO2':  [47.23, 55.15],
                      'TiO2':  [0.74, 0.94],
                      'Al2O3': [14.87, 17.73],
                      'FeO':   [5.08, 7.51],
                      'MgO':   [7.52, 9.26],
                      'CaO':   [11.99, 13.46],
                      'Na2O':  [2.29, 2.67],
                      'K2O':   [1.79, 2.17]
                      }


def crf_stromboli(calibval=None, sample={}):
    return crf_generic(calibval, sample, stromboliCompRange)


crmsg_Comp_pass = ("The sample appears to be similar in composition to the compositional dataset "
                   "for the selected Carbon model of Allison et al. (2019).")
crmsg_Comp_fail = (" These calibration limits were selected based on the minimum and maximum "
                   "values of these oxides (+-5%) in the calibration dataset. As the Allison et "
                   "al. model incorperates no term for compositional dependence, users must take "
                   "extreme care when extrapolating this model to compositions which differ "
                   "significantly from the calibration dataset. These warnings are simply a "
                   "guide;  we suggest that users carefully compare their major element data to"
                   " the calibration dataset to check for suitability ")
crmsg_Comp_description = ("The Allison et al. (2019) Carbon model is defined for 6 different "
                          "alkali compositions.")
crmsg_Temp = ("All calculations for {model_name} are performed at 1200 C "
              "(inputted Temp={param_val:.1f} {units}). Allison et al. (2019) suggest the "
              "results are likely applicable between 1000-1400°C). ")
crmsg_PressureGreater = ("{param_name} ({param_val:.1f} {units}) is less than the lowest P "
                         "experiment in the calibration dataset ({calib_val:.1f} {units}). ")
crmsg_PressureLess = ("{param_name} ({param_val:.1f} {units}) exceeds the upper limit of "
                      "{calib_val:.1f} {units} suggested by Allison et al. (2019) - Their "
                      "spreadsheet would return 7000 bar for this input. ")
crmsg_H2O = ("{param_name} ({param_val:.1f} {units}) is > {param_val:.1f} {units}: this model "
             "does not account for the effect of H$_2$O on volatile solubility. VESIcal allows "
             "you to combine Allison Carbon with a variety of H$_2$O models. ")
crmsg_Between_Temp = ("{param_name} ({param_val:.1f} {units} is outside the recomended ",
                      "temperature range for {model_name} (1000-1400°C). ")


# Create objects for each model location in order to set their calibration ranges
sunset = carbon(model_loc='sunset')
sfvf = carbon(model_loc='sfvf')
erebus = carbon(model_loc='erebus')
vesuvius = carbon(model_loc='vesuvius')
etna = carbon(model_loc='etna')
stromboli = carbon(model_loc='stromboli')

_crs_to_update = sunset.calibration_ranges

cr_oxide_list = []
for ox in sunsetCompRange.keys():
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
                         ox, sunsetCompRange[ox], calibration_checks.crf_Between, 'wt%',
                         'Allison et al. (2019) sunset carbon',
                         fail_msg=calibration_checks.crmsg_BC_fail,
                         pass_msg=calibration_checks.crmsg_BC_pass,
                         description_msg=calibration_checks.crmsg_Between_description))

sunset.set_calibration_ranges(
    _crs_to_update + [
        calibration_checks.CalibrationRange(
            'pressure', 7000, calibration_checks.crf_LessThan, 'bar',
            'Allison et al. (2019) sunset carbon',
            fail_msg=crmsg_PressureLess,
            pass_msg=calibration_checks.crmsg_LessThan_pass),
        calibration_checks.CalibrationRange(
            'pressure', 4071, calibration_checks.crf_GreaterThan, 'bar',
            'Allison et al. (2019) sunset carbon',
            fail_msg=crmsg_PressureGreater,
            pass_msg=calibration_checks.crmsg_GreaterThan_pass),
        calibration_checks.CalibrationRange(
            'sample', None, crf_sunset, None, None,
            fail_msg=crmsg_Comp_fail,
            pass_msg=crmsg_Comp_pass)] + cr_oxide_list)

# SFVF
_crs_to_update = sfvf.calibration_ranges
cr_oxide_list = []
for ox in sfvfCompRange.keys():
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
            ox, sfvfCompRange[ox], calibration_checks.crf_Between, 'wt%',
            'Allison et al. (2019) sfvf carbon',
            fail_msg=calibration_checks.crmsg_BC_fail,
            pass_msg=calibration_checks.crmsg_BC_pass,
            description_msg=calibration_checks.crmsg_Between_description))

sfvf.set_calibration_ranges(_crs_to_update + [
    calibration_checks.CalibrationRange(
        'pressure', 7000, calibration_checks.crf_LessThan, 'bar',
        'Allison et al. (2019) sfvf carbon',
        fail_msg=crmsg_PressureLess,
        pass_msg=calibration_checks.crmsg_LessThan_pass),
    calibration_checks.CalibrationRange(
        'pressure', 4133, calibration_checks.crf_GreaterThan, 'bar',
        'Allison et al. (2019) sfvf carbon',
        fail_msg=crmsg_PressureGreater,
        pass_msg=calibration_checks.crmsg_GreaterThan_pass),
    calibration_checks.CalibrationRange(
        'sample', None, crf_sfvf, None, None,
        fail_msg=crmsg_Comp_fail,
        pass_msg=crmsg_Comp_pass)] + cr_oxide_list)


# Erebus
_crs_to_update = erebus.calibration_ranges
cr_oxide_list = []
for ox in erebusCompRange.keys():
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
            ox, erebusCompRange[ox], calibration_checks.crf_Between, 'wt%',
            'Allison et al. (2019) erebus carbon',
            fail_msg=calibration_checks.crmsg_BC_fail,
            pass_msg=calibration_checks.crmsg_BC_pass,
            description_msg=calibration_checks.crmsg_Between_description))

erebus.set_calibration_ranges(_crs_to_update + [
    calibration_checks.CalibrationRange(
        'pressure', 7000, calibration_checks.crf_LessThan, 'bar',
        'Allison et al. (2019) erebus carbon',
        fail_msg=crmsg_PressureLess,
        pass_msg=calibration_checks.crmsg_LessThan_pass),
    calibration_checks.CalibrationRange(
        'pressure', 4078, calibration_checks.crf_GreaterThan, 'bar',
        'Allison et al. (2019) erebus carbon',
        fail_msg=crmsg_PressureGreater,
        pass_msg=calibration_checks.crmsg_GreaterThan_pass),
    calibration_checks.CalibrationRange(
        'sample', None, crf_erebus, None, None,
        fail_msg=crmsg_Comp_fail,
        pass_msg=crmsg_Comp_pass)] + cr_oxide_list)


_crs_to_update = etna.calibration_ranges
cr_oxide_list = []
for ox in etnaCompRange.keys():
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
            ox, etnaCompRange[ox], calibration_checks.crf_Between, 'wt%',
            'Allison et al. (2019) etna carbon',
            fail_msg=calibration_checks.crmsg_BC_fail,
            pass_msg=calibration_checks.crmsg_BC_pass,
            description_msg=calibration_checks.crmsg_Between_description))


# Etna
etna.set_calibration_ranges(_crs_to_update + [
    calibration_checks.CalibrationRange(
        'pressure', 7000, calibration_checks.crf_LessThan, 'bar',
        'Allison et al. (2019) etna carbon',
        fail_msg=crmsg_PressureLess,
        pass_msg=calibration_checks.crmsg_LessThan_pass),
    calibration_checks.CalibrationRange(
        'pressure', 485, calibration_checks.crf_GreaterThan, 'bar',
        'Allison et al. (2019) etna carbon',
        fail_msg=crmsg_PressureGreater,
        pass_msg=calibration_checks.crmsg_GreaterThan_pass),
    calibration_checks.CalibrationRange(
        'sample', None, crf_etna, None, None,
        fail_msg=crmsg_Comp_fail,
        pass_msg=crmsg_Comp_pass)] + cr_oxide_list)


# Vesuvius
_crs_to_update = vesuvius.calibration_ranges
cr_oxide_list = []
for ox in vesuviusCompRange.keys():
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
            ox, vesuviusCompRange[ox], calibration_checks.crf_Between, 'wt%',
            'Allison et al. (2019) vesuvius carbon',
            fail_msg=calibration_checks.crmsg_BC_fail,
            pass_msg=calibration_checks.crmsg_BC_pass,
            description_msg=calibration_checks.crmsg_Between_description))

vesuvius.set_calibration_ranges(_crs_to_update + [
    calibration_checks.CalibrationRange(
        'pressure', 7000, calibration_checks.crf_LessThan, 'bar',
        'Allison et al. (2019) vesuvius carbon',
        fail_msg=crmsg_PressureLess,
        pass_msg=calibration_checks.crmsg_LessThan_pass),
    calibration_checks.CalibrationRange(
        'pressure', 269, calibration_checks.crf_GreaterThan, 'bar',
        'Allison et al. (2019) vesuvius carbon',
        fail_msg=crmsg_PressureGreater,
        pass_msg=calibration_checks.crmsg_GreaterThan_pass),
    calibration_checks.CalibrationRange(
        'sample', None, crf_vesuvius, None, None,
        fail_msg=crmsg_Comp_fail,
        pass_msg=crmsg_Comp_pass)] + cr_oxide_list)


# Stromboli
_crs_to_update = stromboli.calibration_ranges
cr_oxide_list = []
for ox in stromboliCompRange.keys():
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
            ox, stromboliCompRange[ox], calibration_checks.crf_Between, 'wt%',
            'Allison et al. (2019) stromboli carbon',
            fail_msg=calibration_checks.crmsg_BC_fail,
            pass_msg=calibration_checks.crmsg_BC_pass,
            description_msg=calibration_checks.crmsg_Between_description))

stromboli.set_calibration_ranges(_crs_to_update + [
    calibration_checks.CalibrationRange(
        'pressure', 7000, calibration_checks.crf_LessThan, 'bar',
        'Allison et al. (2019) stromboli carbon',
        fail_msg=crmsg_PressureLess,
        pass_msg=calibration_checks.crmsg_LessThan_pass),
    calibration_checks.CalibrationRange(
        'pressure', 524, calibration_checks.crf_GreaterThan, 'bar',
        'Allison et al. (2019) stromboli carbon',
        fail_msg=crmsg_PressureGreater,
        pass_msg=calibration_checks.crmsg_GreaterThan_pass),
    calibration_checks.CalibrationRange(
        'sample', None, crf_stromboli, None, None,
        fail_msg=crmsg_Comp_fail,
        pass_msg=crmsg_Comp_pass)] + cr_oxide_list)



from VESIcal import core
from VESIcal.thermo import thermo_core as tc


def calculate_liquid_density(sample, pressure, temperature, **kwargs):
    """ Calculates the density of the given liquid using the DensityX model of
    Iacovino and Till (2019).

    Parameters
    ----------
    sample  Sample class
        Magma major element composition.
    pressure    float
        Total pressure in bars.
    temperature     float
        Temperature in degrees C

    Returns
    -------
    float
        The density of the liquid in g/L, rounded to 3 dp.
    """

    comp_molfrac = sample.get_composition(units='mol_oxides')
    # get only species considered in density calculation
    comp_molfrac = {k: v for k, v in comp_molfrac.items() if
                    k in tc.densitySpecies}
    tempK = temperature + 273.15

    indiv_Vliqs = []
    for k, v in comp_molfrac.items():
        indiv_Vliq = ((tc.MolarVolumes[k] + (tc.dVdTs[k] * (tempK-tc.Trefs[k])) +
                      (tc.dVdPs[k] * (pressure-1))) * v)
        indiv_Vliqs.append(indiv_Vliq)

    Vliq_sum = sum(indiv_Vliqs)
    Sum_X_MW = sum([v*core.oxideMass[k] for k, v in comp_molfrac.items()])

    density_g_per_cm3 = Sum_X_MW/Vliq_sum
    density_g_per_L = density_g_per_cm3*1000

    return round(density_g_per_L, 3)



from abc import abstractmethod
import warnings as w

from VESIcal.thermo import densityx, giordano


class ThermoCalculate(object):
    """
    The ThermoCalculate object is designed to emulate the VESIcal Calculate
    object. It provides a common workflow for all additional thermodynamic
    calculations one might wish to perform on a sample. It is separate from
    the VESIcal Calculate class as that is specifically for volatile
    solubility models
    """
    def __init__(self, sample, silence_warnings=False, **kwargs):
        """
        Initializes the calculation.

        Parameters
        ----------
        sample:    Sample class
            The rock composition as a Sample object.
        silence_warnings:     bool
            Silence warnings about calibration ranges. Default is False.
        """
        self.sample = sample

        self.result = self.calculate(sample=self.sample, **kwargs)
        self.calib_check = self.check_calibration_range(sample=self.sample,
                                                        **kwargs)

        if self.calib_check is not None and silence_warnings is False:
            if self.calib_check != '':
                w.warn(self.calib_check, RuntimeWarning)

    @abstractmethod
    def calculate(self):
        """ """

    @abstractmethod
    def check_calibration_range(self):
        """ """


class calculate_liquid_density(ThermoCalculate):
    """
    Calculates the density of the liquid using the DensityX model. Using
    this interface will preprocess the sample, run the calculation, and then
    check the calibration ranges. All parameters required by the chosen model
    must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure:   float
        Total pressure in bars.
    temperature:  float
        Temperature in degrees C.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        Density in g/L.
    """

    def calculate(self, sample, pressure, temperature, **kwargs):
        density_g_per_L = densityx.calculate_liquid_density(sample=sample,
                                                            pressure=pressure,
                                                            temperature=temperature,
                                                            **kwargs)
        return density_g_per_L

    def check_calibration_range(self, sample, pressure, temperature, **kwargs):
        pass


class calculate_liquid_viscosity(ThermoCalculate):
    """
    Calculates the density of the liquid using the Giordano et al. (2008)
    model. Using this interface will preprocess the sample, run the calculation,
    and then check the calibration ranges. All parameters required by the chosen
    model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    temperature:  float
        Temperature in degrees C.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        Log viscosity in Pa*s.
    """

    def calculate(self, sample, temperature, **kwargs):
        viscosity_log_nu_Pas = giordano.calculate_liquid_viscosity(sample=sample,
                                                                   temperature=temperature,
                                                                   **kwargs)
        return viscosity_log_nu_Pas

    def check_calibration_range(self, sample, temperature, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: tasplot.py
 
############################################################################
#
#  COPYRIGHT:  (C) 2015 John A Stevenson / @volcan01010
#                       Joaquin Cortés
#  WEBSITE: http://all-geo.org/volcan01010
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  http://www.gnu.org/licenses/gpl-3.0.html
#
#############################################################################/

__all__ = ['add_LeMaitre_fields']

class MissingModuleException(Exception):
    pass

 
# Plot LeMaitre lines
def add_LeMaitre_fields(plot_axes, fontsize=8, color=(0.6, 0.6, 0.6)):
    """Add fields for geochemical classifications from LeMaitre et al (2002)
    to pre-existing axes.  If necessary, the axes object can be retrieved via
    plt.gca() command. e.g.
    
    ax1 = plt.gca()
    add_LeMaitre_fields(ax1)
    ax1.plot(silica, total_alkalis, 'o')
    
    Fontsize and color options can be used to change from the defaults.
    
    It may be necessary to follow the command with plt.draw() to update
    the plot.
    
    Le Maitre RW (2002) Igneous rocks : IUGS classification and glossary of
        terms : recommendations of the International Union of Geological 
        Sciences Subcommission on the Systematics of igneous rocks, 2nd ed. 
        Cambridge University Press, Cambridge
"""

    # Check matplotlib is imported
    import sys
    if 'matplotlib.pyplot' not in sys.modules:
        raise MissingModuleException("""Matplotlib not imported.
        Matplotlib is installed as part of many scientific packages and is
        required to create plots.""")
    
    # Check that plot_axis can plot
    if 'plot' not in dir(plot_axes):
        raise TypeError('plot_axes is not a matplotlib axes instance.')
    
    # Prepare the field information
    from collections import namedtuple
    FieldLine = namedtuple('FieldLine', 'x1 y1 x2 y2')
    lines = (FieldLine(x1=41, y1=0, x2=41, y2=7),
             FieldLine(x1=41, y1=7, x2=52.5, y2=14),
             FieldLine(x1=45, y1=0, x2=45, y2=5),
             FieldLine(x1=41, y1=3, x2=45, y2=3),
             FieldLine(x1=45, y1=5, x2=61, y2=13.5),
             FieldLine(x1=45, y1=5, x2=52, y2=5),
             FieldLine(x1=52, y1=5, x2=69, y2=8),
             FieldLine(x1=49.4, y1=7.3, x2=52, y2=5),
             FieldLine(x1=52, y1=5, x2=52, y2=0),
             FieldLine(x1=48.4, y1=11.5, x2=53, y2=9.3),
             FieldLine(x1=53, y1=9.3, x2=57, y2=5.9),
             FieldLine(x1=57, y1=5.9, x2=57, y2=0),
             FieldLine(x1=52.5, y1=14, x2=57.6, y2=11.7),
             FieldLine(x1=57.6, y1=11.7, x2=63, y2=7),
             FieldLine(x1=63, y1=7, x2=63, y2=0),
             FieldLine(x1=69, y1=12, x2=69, y2=8),
             FieldLine(x1=45, y1=9.4, x2=49.4, y2=7.3),
             FieldLine(x1=69, y1=8, x2=77, y2=0))

    FieldName = namedtuple('FieldName', 'name x y rotation')
    names = (FieldName('Picro\nbasalt', 43, 2, 0),
             FieldName('Basalt', 48.5, 2, 0),
             FieldName('Basaltic\nandesite', 54.5, 2, 0),
             FieldName('Andesite', 60, 2, 0),
             FieldName('Dacite', 68.5, 2, 0),
             FieldName('Rhyolite', 76, 9, 0),
             FieldName('Trachyte\n(Q < 20%)\n\nTrachydacite\n(Q > 20%)',
                       64.5, 11.5, 0),
             FieldName('Basaltic\ntrachyandesite', 53, 8, -20),
             FieldName('Trachy-\nbasalt', 49, 6.2, 0),
             FieldName('Trachyandesite', 57.2, 9, 0),
             FieldName('Phonotephrite', 49, 9.6, 0),
             FieldName('Tephriphonolite', 53.0, 11.8, 0),
             FieldName('Phonolite', 57.5, 13.5, 0),
             FieldName('Tephrite\n(Ol < 10%)', 45, 8, 0),
             FieldName('Foidite', 44, 11.5, 0),
             FieldName('Basanite\n(Ol > 10%)', 43.5, 6.5, 0))

    # Plot the lines and fields
    for line in lines:
        plot_axes.plot([line.x1, line.x2], [line.y1, line.y2],
                       '-', color=color, zorder=0)
    for name in names:
        plot_axes.text(name.x, name.y, name.name, color=color, size=fontsize,
                 horizontalalignment='center', verticalalignment='top',
                 rotation=name.rotation, zorder=0)


from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import model_classes
from VESIcal import sample_class
from VESIcal import vplot
from VESIcal import batchfile  # needed for status_bar functions

from copy import deepcopy
import numpy as np
import pandas as pd
import warnings as w
import sys
from contextlib import redirect_stdout
import io

from thermoengine import equilibrate

# Variable to send H2O driver warnings into the void
_f = io.StringIO()

w.filterwarnings("ignore", message="rubicon.objc.ctypes_patch has only been tested ")

# -------------- MELTS preamble --------------- #
# instantiate thermoengine equilibrate MELTS instance
melts = equilibrate.MELTSmodel("1.2.0")

# Suppress phases not required in the melts simulation
phases = melts.get_phase_names()
for phase in phases:
    melts.set_phase_inclusion_status({phase: False})
melts.set_phase_inclusion_status({"Fluid": True, "Liquid": True})
# --------------------------------------------- #


class MagmaSat(model_classes.Model):
    """
    An object to instantiate a thermoengine equilibrate class
    """

    def __init__(self):
        self.melts_version = (
            "1.2.0"  # just here so users can see which version is being used
        )

        self.set_volatile_species(["H2O", "CO2"])
        self.set_calibration_ranges(
            [
                calibration_checks.CalibrationRange(
                    "pressure",
                    [0.0, 20000.0],
                    calibration_checks.crf_Between,
                    "bar",
                    "MagmaSat",
                    fail_msg=calibration_checks.crmsg_Between_fail,
                    pass_msg=calibration_checks.crmsg_Between_pass,
                    description_msg=calibration_checks.crmsg_Between_description,
                ),
                calibration_checks.CalibrationRange(
                    "temperature",
                    [800, 1400],
                    calibration_checks.crf_Between,
                    "oC",
                    "MagmaSat",
                    fail_msg=calibration_checks.crmsg_Between_fail,
                    pass_msg=calibration_checks.crmsg_Between_pass,
                    description_msg=calibration_checks.crmsg_Between_description,
                ),
            ]
        )
        self.model_type = "MagmaSat"

    def preprocess_sample(self, sample):
        """
        Returns sample with 0.0 values for any oxides not passed.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        Returns
        -------
        Sample class object

        """
        _sample = deepcopy(sample)
        _sample = _sample.get_composition(
            units="wtpt_oxides",
            normalization=_sample.default_normalization,
            asSampleClass=True,
        )

        # set any necessary magmasat oxides to 0 if no value given
        for oxide in core.magmasat_oxides:
            if oxide in _sample.get_composition():
                pass
            else:
                _sample.change_composition({oxide: 0.0})

        # remove any non-magmasat oxides
        for oxide in _sample.get_composition().index:
            if oxide in core.magmasat_oxides:
                pass
            else:
                _sample.delete_oxide(oxide)

        self.bulk_comp_orig = _sample.get_composition(units="wtpt_oxides")

        return _sample

    def check_calibration_range(self, parameters, **kwargs):
        """Checks whether supplied parameters and calculated results are within the calibration
        range of the model, defined by the CalibrationRange objects. An empty string will be
        returned if all parameters are within the calibration range. If a parameter is not within
        the calibration range, a description of the problem will be returned in the string.

        Parameters
        ----------
        parameters     dict
            Dictionary keys are the names of the parameters to be checked, e.g., pressure
            temperature, SiO2, etc. Values are the values of each parameter. A complete set
            need not be given.

        Returns
        -------
        str
            String description of any parameters falling outside of the calibration range.
        """
        s = ""
        for cr in self.calibration_ranges:
            if cr.check(parameters) is False:
                s += cr.string(parameters, report_nonexistance=False)
            if "notsaturated" in kwargs:
                s += "Sample not saturated at these conditions."
        return s

    def get_calibration_range(self):
        """Returns a string describing the calibration ranges defined by the CalibrationRange
        objects for the model.

        Returns
        -------
        str
            String description of the calibration range objects."""
        s = ""
        for cr in self.calibration_ranges:
            s += cr.string(None)
        return s

    def get_fluid_mass(self, sample, temperature, pressure, H2O, CO2):
        """An internally used function to calculate fluid mass.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature: float
            Temperature in degrees C.

        pressure: float
            Pressure in bars

        H2O: float
            wt% H2O in the system

        CO2: float
            wt% CO2 in the system

        Returns
        -------
        float
            mass of the fluid in grams
        """
        pressureMPa = pressure / 10.0

        bulk_comp_dict = sample.get_composition(units="wtpt_oxides")
        bulk_comp_dict = {oxide: bulk_comp_dict[oxide] for oxide in
                          core.magmasat_oxides}
        bulk_comp_dict["H2O"] = H2O
        bulk_comp_dict["CO2"] = CO2
        melts.set_bulk_composition(bulk_comp_dict)

        with redirect_stdout(_f):
            output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

        return fluid_mass

    def get_XH2O_fluid(self, sample, temperature, pressure, H2O, CO2):
        """An internally used function to calculate fluid composition.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature: float
            Temperature in degrees C.

        pressure: float
            Pressure in bars

        H2O: float
            wt% H2O in the system

        CO2: float
            wt% CO2 in the system

        Returns
        -------
        float
            Mole fraction of H2O in the H2O-CO2 fluid

        """
        _sample = self.preprocess_sample(sample)

        pressureMPa = pressure / 10.0

        _sample.change_composition({"H2O": H2O, "CO2": CO2})
        melts.set_bulk_composition(
            _sample.get_composition(units="wtpt_oxides", normalization="none")
        )

        with redirect_stdout(_f):
            output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_comp = melts.get_composition_of_phase(
            xmlout, phase_name="Fluid", mode="component"
        )
        # NOTE mode='component' returns endmember component keys with values in mol fraction.

        if "Water" in fluid_comp:
            H2O_fl = fluid_comp["Water"]
        else:
            H2O_fl = 0.0

        return H2O_fl

    def calculate_dissolved_volatiles(
        self,
        sample,
        temperature,
        pressure,
        X_fluid=1,
        H2O_guess=0.0,
        verbose=False,
        **kwargs
    ):
        """
        Calculates the amount of H2O and CO2 dissolved in a magma at saturation at the given P/T
        conditions and fluid composition. Fluid composition will be matched to within 0.0001 mole
        fraction.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature: float or int
            Temperature, in degrees C.

        presure: float or int
            Pressure, in bars.

        X_fluid: float or int
            The default value is 1. The mole fraction of H2O in the H2O-CO2 fluid. X_fluid=1 is a
            pure H2O fluid. X_fluid=0 is a pure CO2 fluid.

        verbose: bool
            OPTIONAL: Default is False. If set to True, returns H2O and CO2 concentration in the
            melt, H2O and CO2 concentration in the fluid, mass of the fluid in grams, and
            proportion of fluid in the system in wt%.

        Returns
        -------
        dict
            A dictionary of dissolved volatile concentrations in wt% with keys H2O and CO2.
        """
        _sample = self.preprocess_sample(sample)

        if isinstance(X_fluid, int) or isinstance(X_fluid, float):
            pass
        else:
            raise core.InputError("X_fluid must be type int or float")

        if isinstance(H2O_guess, int) or isinstance(H2O_guess, float):
            pass
        else:
            raise core.InputError("H2O_guess must be type int or float")

        pressureMPa = pressure / 10.0

        if X_fluid != 0 and X_fluid != 1:
            if X_fluid < 0.001 or X_fluid > 0.999:
                raise core.InputError(
                    "X_fluid is calculated to a precision of 0.0001 mole "
                    "fraction. Value for X_fluid must be between 0.0001 and "
                    "0.9999."
                )

        H2O_val = H2O_guess
        CO2_val = 0.0
        fluid_mass = 0.0
        while fluid_mass <= 0:
            if X_fluid == 0:
                CO2_val += 0.1
            elif X_fluid >= 0.5:
                H2O_val += 0.2
                # NOTE this is setting XH2Owt of the system (not of the fluid) to X_fluid
                CO2_val = (H2O_val / X_fluid) - H2O_val
                # TODO this is what needs to be higher for higher XH2O. Slows down computation
                # by a second or two
            else:
                H2O_val += 0.1
                # NOTE this is setting XH2Owt of the system (not of the fluid) to X_fluid
                CO2_val = (H2O_val / X_fluid) - H2O_val
                # TODO this is what needs to be higher for higher XH2O. Slows down computation
                # by a second or two

            fluid_mass = self.get_fluid_mass(
                _sample, temperature, pressure, H2O_val, CO2_val
            )

        _sample.change_composition(
            {"H2O": H2O_val, "CO2": CO2_val}, units="wtpt_oxides"
        )
        melts.set_bulk_composition(
            _sample.get_composition(units="wtpt_oxides", normalization="none")
        )

        with redirect_stdout(_f):
            output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        liquid_comp = melts.get_composition_of_phase(
            xmlout, phase_name="Liquid", mode="oxide_wt"
        )
        fluid_comp = melts.get_composition_of_phase(
            xmlout, phase_name="Fluid", mode="component"
        )

        if "Water" in fluid_comp:
            H2O_fl = fluid_comp["Water"]
        else:
            H2O_fl = 0.0

        XH2O_fluid = H2O_fl

        # ------ Coarse Check ------ #
        while XH2O_fluid < X_fluid - 0.1:  # too low coarse check
            H2O_val += 0.2
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        while XH2O_fluid > X_fluid + 0.1:  # too high coarse check
            CO2_val += 0.1
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        # ------ Refinement 1 ------ #
        while XH2O_fluid < X_fluid - 0.01:  # too low refinement 1
            H2O_val += 0.05
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        while XH2O_fluid > X_fluid + 0.01:  # too high refinement 1
            CO2_val += 0.01
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        # ------ Refinement 2 ------ #
        while XH2O_fluid < X_fluid - 0.001:  # too low refinement 2
            H2O_val += 0.005
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        while XH2O_fluid > X_fluid + 0.001:  # too high refinement 2
            CO2_val += 0.001
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        # ------ Final refinement ------ #
        while XH2O_fluid < X_fluid - 0.0001:  # too low final refinement
            H2O_val += 0.001
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        while XH2O_fluid > X_fluid + 0.0001:  # too high final refinement
            CO2_val += 0.0001
            XH2O_fluid = self.get_XH2O_fluid(
                sample, temperature, pressure, H2O_val, CO2_val
            )

        # ------ Get calculated values ------ #
        _sample.change_composition(
            {"H2O": H2O_val, "CO2": CO2_val}, units="wtpt_oxides"
        )
        melts.set_bulk_composition(
            _sample.get_composition(units="wtpt_oxides", normalization="none")
        )

        with redirect_stdout(_f):
            output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")
        system_mass = melts.get_mass_of_phase(xmlout, phase_name="System")
        liquid_comp = melts.get_composition_of_phase(
            xmlout, phase_name="Liquid", mode="oxide_wt"
        )
        fluid_comp = melts.get_composition_of_phase(
            xmlout, phase_name="Fluid", mode="component"
        )

        if "H2O" in liquid_comp:
            H2O_liq = liquid_comp["H2O"]
        else:
            H2O_liq = 0

        if "CO2" in liquid_comp:
            CO2_liq = liquid_comp["CO2"]
        else:
            CO2_liq = 0

        if "Water" in fluid_comp:
            H2O_fl = fluid_comp["Water"]
        else:
            H2O_fl = 0.0
        if "Carbon Dioxide" in fluid_comp:
            CO2_fl = fluid_comp["Carbon Dioxide"]
        else:
            CO2_fl = 0.0

        XH2O_fluid = H2O_fl

        if verbose:
            return {
                "temperature": temperature,
                "pressure": pressure,
                "H2O_liq": H2O_liq,
                "CO2_liq": CO2_liq,
                "XH2O_fl": H2O_fl,
                "XCO2_fl": CO2_fl,
                "FluidProportion_wt": 100 * fluid_mass / system_mass,
            }

        if verbose is False:
            return {"CO2_liq": CO2_liq, "H2O_liq": H2O_liq}

    def calculate_equilibrium_fluid_comp(
        self, sample, temperature, pressure, verbose=False, **kwargs
    ):
        """
        Returns H2O and CO2 concentrations in wt% in a fluid in equilibrium with the given sample
        at the given P/T condition.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature: float or int
            Temperature, in degrees C.

        presure: float or int
            Pressure, in bars.

        verbose: bool
            OPTIONAL: Default is False. If set to True, returns H2O and CO2 concentration in the
            fluid, mass of the fluid in grams, and proportion of fluid in the system in wt%.

        Returns
        -------
        dict
            A dictionary of fluid composition in wt% with keys 'H2O' and 'CO2' is returned.
        """
        _sample = self.preprocess_sample(sample)
        bulk_comp_dict = _sample.get_composition(
            units="wtpt_oxides", normalization="none"
        )

        if isinstance(temperature, float) or isinstance(temperature, int):
            pass
        else:
            raise core.InputError("temp must be type float or int")

        if isinstance(pressure, float) or isinstance(pressure, int):
            pass
        else:
            raise core.InputError("presure must be type float or int")

        # Check if only single volatile species is passed. If so, can skip calculations.
        if bulk_comp_dict["H2O"] == 0:
            if bulk_comp_dict["CO2"] == 0:
                if verbose is False:
                    return {"CO2": 0.0, "H2O": 0.0}
                if verbose:
                    return {
                        "CO2": 0.0,
                        "H2O": 0.0,
                        "FluidMass_grams": 0.0,
                        "FluidProportion_wt": 0.0,
                    }
            else:
                if verbose is False:
                    return {"CO2": 1.0, "H2O": 0.0}
        else:
            if bulk_comp_dict["CO2"] == 0:
                if verbose is False:
                    return {"CO2": 0.0, "H2O": 1.0}

        pressureMPa = pressure / 10.0

        melts.set_bulk_composition(bulk_comp_dict)

        with redirect_stdout(_f):
            output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")
        flsystem_wtper = (
            100
            * fluid_mass
            / (fluid_mass + melts.get_mass_of_phase(xmlout, phase_name="Liquid"))
        )

        if fluid_mass > 0.0:
            fluid_comp = melts.get_composition_of_phase(
                xmlout, phase_name="Fluid", mode="component"
            )
            fluid_comp_H2O = fluid_comp["Water"]
            fluid_comp_CO2 = fluid_comp["Carbon Dioxide"]
        else:
            fluid_comp_H2O = 0
            fluid_comp_CO2 = 0

        melts.set_bulk_composition(self.bulk_comp_orig)  # reset

        if verbose is False:
            return {"CO2": fluid_comp_CO2, "H2O": fluid_comp_H2O}

        if verbose:
            return {
                "CO2": fluid_comp_CO2,
                "H2O": fluid_comp_H2O,
                "FluidMass_grams": fluid_mass,
                "FluidProportion_wt": flsystem_wtper,
            }

    def calculate_saturation_pressure(
        self, sample, temperature, verbose=False, **kwargs
    ):
        """
        Calculates the saturation pressure of a sample composition.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature: flaot or int
            Temperature of the sample in degrees C.

        verbose: bool
            OPTIONAL: Default is False. If set to False, only the saturation pressure is returned.
            If set to True, the saturation pressure, mass of fluid in grams, proportion of fluid
            in wt%, and H2O and CO2 concentrations in the fluid in mole fraction are all returned
            in a dict.

        Returns
        -------
        float or dict
            If verbose is set to False: Saturation pressure in bars.
            If verbose is set to True: dict of all calculated values.
        """
        _sample = self.preprocess_sample(sample)

        melts.set_bulk_composition(
            _sample.get_composition(units="wtpt_oxides", normalization="fixedvolatiles")
        )
        # Coarse search
        # NOTE that pressure is in MPa for MagmaSat calculations but reported in bars.
        pressureMPa = 2000

        # Check if saturated at 2000 MPa (rare, for deep samples)
        with redirect_stdout(_f):
            output = melts.equilibrate_tp(temperature, pressureMPa, initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

        if fluid_mass <= 0:  # if not sat'd at 2000 MPa
            while fluid_mass <= 0:
                pressureMPa -= 100
                if pressureMPa <= 0:
                    break

                with redirect_stdout(_f):
                    output = melts.equilibrate_tp(
                        temperature, pressureMPa, initialize=True
                    )
                (status, temperature, pressureMPa, xmlout) = output[0]
                fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

            fluid_mass = 0
            pressureMPa += 100

        elif fluid_mass > 0:  # if sat'd at 2000 MPa, add pressure
            while fluid_mass > 0:
                pressureMPa += 100

                melts.set_bulk_composition(
                    _sample.get_composition(
                        units="wtpt_oxides", normalization="fixedvolatiles"
                    )
                )
                with redirect_stdout(_f):
                    output = melts.equilibrate_tp(
                        temperature, pressureMPa, initialize=True
                    )
                (status, temperature, pressureMPa, xmlout) = output[0]
                fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

            fluid_mass = 1.0
            pressureMPa -= 100

        # Refined search 1
        melts.set_bulk_composition(
            _sample.get_composition(units="wtpt_oxides", normalization="fixedvolatiles")
        )

        if fluid_mass <= 0:  # proceed down pressure search
            while fluid_mass <= 0:
                pressureMPa -= 10
                if pressureMPa <= 0:
                    break

                with redirect_stdout(_f):
                    output = melts.equilibrate_tp(
                        temperature, pressureMPa, initialize=True
                    )
                (status, temperature, pressureMPa, xmlout) = output[0]
                fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

            fluid_mass = 0
            pressureMPa += 10

        elif fluid_mass > 0:  # proceed upward pressure search
            while fluid_mass > 0:
                pressureMPa += 10

                melts.set_bulk_composition(
                    _sample.get_composition(
                        units="wtpt_oxides", normalization="fixedvolatiles"
                    )
                )
                with redirect_stdout(_f):
                    output = melts.equilibrate_tp(
                        temperature, pressureMPa, initialize=True
                    )
                (status, temperature, pressureMPa, xmlout) = output[0]
                fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

            fluid_mass = 1.0
            pressureMPa -= 10

        # Refined search 2
        melts.set_bulk_composition(
            _sample.get_composition(units="wtpt_oxides", normalization="fixedvolatiles")
        )

        if fluid_mass <= 0:  # proceed down pressure search
            while fluid_mass <= 0:
                pressureMPa -= 1
                if pressureMPa <= 0:
                    break

                with redirect_stdout(_f):
                    output = melts.equilibrate_tp(
                        temperature, pressureMPa, initialize=True
                    )
                (status, temperature, pressureMPa, xmlout) = output[0]
                fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

        elif fluid_mass > 0:  # proceed upward pressure search
            while fluid_mass > 0:
                pressureMPa += 1

                melts.set_bulk_composition(
                    _sample.get_composition(
                        units="wtpt_oxides", normalization="fixedvolatiles"
                    )
                )
                with redirect_stdout(_f):
                    output = melts.equilibrate_tp(
                        temperature, pressureMPa, initialize=True
                    )
                (status, temperature, pressureMPa, xmlout) = output[0]
                fluid_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")

        if pressureMPa != np.nan:
            satP = pressureMPa * 10  # convert pressure to bars
            flmass = fluid_mass
            flsystem_wtper = (
                100
                * fluid_mass
                / (fluid_mass + melts.get_mass_of_phase(xmlout, phase_name="Liquid"))
            )
            flcomp = melts.get_composition_of_phase(
                xmlout, phase_name="Fluid", mode="component"
            )
            try:
                flH2O = flcomp["Water"]
            except Exception:
                flH2O = 0.0
            try:
                flCO2 = flcomp["Carbon Dioxide"]
            except Exception:
                flCO2 = 0.0
        else:
            flmass = np.nan
            flsystem_wtper = np.nan
            flH2O = np.nan
            flCO2 = np.nan
            warnmessage = "Calculation failed."

        melts.set_bulk_composition(
            self.bulk_comp_orig
        )  # this needs to be reset always!

        if verbose is False:
            try:
                w.warn(warnmessage)
            except Exception:
                pass
            return satP

        elif verbose:
            try:
                w.warn(warnmessage)
            except Exception:
                pass
            return {
                "SaturationP_bars": satP,
                "FluidMass_grams": flmass,
                "FluidProportion_wt": flsystem_wtper,
                "XH2O_fl": flH2O,
                "XCO2_fl": flCO2,
            }

    def calculate_isobars_and_isopleths(
        self,
        sample,
        temperature,
        pressure_list,
        isopleth_list=None,
        smooth_isobars=True,
        smooth_isopleths=True,
        print_status=True,
        **kwargs
    ):
        """
        Calculates isobars and isopleths at a constant temperature for a given sample. Isobars can
        be calculated for any number of pressures. Isobars are calculated using 5 XH2O values
        (0, 0.25, 0.5, 0.75, 1).

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature: float
            Temperature in degrees C.

        pressure_list: list or float
            List of all pressure values at which to calculate isobars, in bars. If only one value
            is passed it can be as float instead of list.

        isopleth_list: list or float
            OPTIONAL: Default value is None in which case only isobars will be calculated.
            List of all fluid compositions in mole fraction H2O (XH2Ofluid) at which to calcualte
            isopleths. Values can range from 0-1. If only one value is passed it can be as float
            instead of list.

        smooth_isobars: bool
            OPTIONAL. Default is True. If set to True, polynomials will be fit to the computed
            isobar points.

        smooth_isopleths: bool
            OPTIONAL. Default is True. If set to True, polynomials will be fit to the computed
            isopleth points.

        print_status: bool
            OPTIONAL: Default is True. If set to True, progress of the calculations will be
            printed to the terminal.

        Returns
        -------
        pandas DataFrame objects
            Two pandas DataFrames are returned; the first has isobar data, and the second has
            isopleth data. Columns in the isobar dataframe are 'Pressure', 'H2Omelt', and
            'CO2melt', correpsonding to pressure in bars and dissolved H2O and CO2 in the liquid
            in wt%. Columns in the isopleth dataframe are 'Pressure', 'H2Ofl', and 'CO2fl',
            corresponding to pressure in bars and H2O and CO2 concentration in the H2O-CO2 fluid,
            in wt%.
        """

        if isinstance(pressure_list, list):
            P_vals = pressure_list
        elif isinstance(pressure_list, int) or isinstance(pressure_list, float):
            P_vals = [pressure_list]
        else:
            raise core.InputError("pressure_list must be a single float (1000.0), int (1000), or "
                                  "list of those [1000, 2000.0, 3000].")

        if isopleth_list is None:
            pass
        elif isinstance(isopleth_list, list):
            iso_vals = isopleth_list
        else:
            iso_vals = [isopleth_list]

        _sample = self.preprocess_sample(sample)

        required_iso_vals = [0, 0.25, 0.5, 0.75, 1]
        all_iso_vals = iso_vals + required_iso_vals
        all_iso_vals = list(dict.fromkeys(all_iso_vals))  # remove duplicates
        all_iso_vals.sort()  # sort from smallest to largest

        isobar_data = []
        isopleth_data = []
        for X in iso_vals:
            isopleth_data.append([X, 0.0, 0.0])
        # Calculate equilibrium phase assemblage for all P/T conditions, check if saturated in
        # fluid...
        for i in P_vals:
            guess = 0.0
            if print_status:
                print("Calculating isobar at " + str(i) + " bars")
            X_iter = 0
            for X in all_iso_vals:
                X_iter += 1
                if print_status:
                    if isopleth_list is not None and X in iso_vals:
                        sys.stdout.write(
                            "\r Calculating isopleth at XH2Ofluid = "
                            + str(X)
                            + "               "
                        )
                    if X not in iso_vals:
                        sys.stdout.write(
                            "\r Calculating isobar control point at XH2Ofluid = "
                            + str(X)
                            + "               "
                        )
                    if X_iter == len(all_iso_vals):
                        sys.stdout.write(
                            "\r done.                                               "
                            "                                                       "
                            "                     \n"
                        )
                saturated_vols = self.calculate_dissolved_volatiles(
                    sample=_sample,
                    temperature=temperature,
                    pressure=i,
                    H2O_guess=guess,
                    X_fluid=X,
                )

                if X in required_iso_vals:
                    isobar_data.append([i, saturated_vols["H2O_liq"], saturated_vols["CO2_liq"]])
                if X in iso_vals:
                    isopleth_data.append([X, saturated_vols["H2O_liq"], saturated_vols["CO2_liq"]])

                guess = saturated_vols["H2O_liq"]

        if print_status:
            print("Done!")

        isobars_df = pd.DataFrame(isobar_data, columns=["Pressure", "H2O_liq", "CO2_liq"])
        isopleths_df = pd.DataFrame(isopleth_data, columns=["XH2O_fl", "H2O_liq", "CO2_liq"])

        melts.set_bulk_composition(self.bulk_comp_orig)  # reset

        if smooth_isobars:
            isobars_smoothed = vplot.smooth_isobars_and_isopleths(isobars=isobars_df)
            res_isobars = isobars_smoothed.copy()
        else:
            res_isobars = isobars_df.copy()

        if smooth_isopleths:
            isopleths_smoothed = vplot.smooth_isobars_and_isopleths(isopleths=isopleths_df)
            res_isopleths = isopleths_smoothed.copy()
        else:
            res_isopleths = isopleths_df.copy()

        return res_isobars, res_isopleths

    def calculate_degassing_path(self, sample, temperature, pressure="saturation",
                                 fractionate_vapor=0.0, init_vapor=0.0, steps=50, **kwargs):
        """
        Calculates degassing path for one sample

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

            Legacy info follows, might still be useful?
            If pulling from an uploaded file
            with data for many samples, first call get_sample_composition() to get the sample
            desired. Then pass the result into this function.

        temperature: float
            Temperature at which to calculate degassing paths, in degrees C.

        pressure: float
            OPTIONAL. The perssure at which to begin the degassing calculations. Default value is
            'saturation', which runs the calculation with the initial pressure at the saturation
            pressure. If a pressure greater than the saturation pressure is input, the calculation
            will start at saturation, since this is the first pressure at which any degassing will
            occur.

        fractionate_vapor: float
            OPTIONAL. Proportion of vapor removed at each pressure step.
            Default value is 0.0 (completely closed-system degassing). Specifies the type of
            calculation performed, either closed system (0.0) or open system (1.0) degassing. If
            any value between <1.0 is chosen, user can also specify the 'init_vapor' argument
            (see below). A value in between 0 and 1 will remove that proportion of vapor at each
            step. For example, for a value of 0.2, the calculation will remove 20% of the vapor
            and retain 80% of the vapor at each pressure step.

        init_vapor: float
            OPTIONAL. Default value is 0.0. Specifies the amount of vapor (in wt%) coexisting
            with the melt before degassing.

        steps: int
            OPTIONAL. Default value is 50. Specifies the number of steps in pressure space at
            which dissolved volatile concentrations are calculated.

        Returns
        -------
        pandas DataFrame object

        """
        sys.stdout.write("Finding saturation point... ")  # print start of calculation to terminal
        _sample = self.preprocess_sample(sample)

        # Normalize sample composition
        _normed_comp = _sample.get_composition(normalization="standard")
        _sample.change_composition(_normed_comp)
        _sample_dict = _sample.get_composition()

        # ------ RESET MELTS ------ #
        # MELTS needs to be reloaded here. If an unfeasible composition gets set inside of MELTS,
        # which can happen when running open-system degassing path calcs, the following calls to
        # MELTS will fail. This prevents that from happening.
        melts = equilibrate.MELTSmodel("1.2.0")

        # Suppress phases not required in the melts simulation
        phases = melts.get_phase_names()
        for phase in phases:
            melts.set_phase_inclusion_status({phase: False})
        melts.set_phase_inclusion_status({"Fluid": True, "Liquid": True})

        melts.set_bulk_composition(_sample_dict)
        # ------------------------- #

        # Get saturation pressure
        data = self.calculate_saturation_pressure(sample=_sample, temperature=temperature,
                                                  verbose=True)

        if pressure == "saturation" or pressure >= data["SaturationP_bars"]:
            SatP_MPa = data["SaturationP_bars"] / 10.0
        else:
            SatP_MPa = pressure / 10.0

        # convert number of steps to step size
        MPa_step = SatP_MPa / steps
        if MPa_step < 1:
            MPa_step = 1

        P_array = np.arange(1.0, SatP_MPa, MPa_step)
        P_array = -np.sort(-P_array)
        # add last few MPa steps
        P_array = np.append(P_array, 0.5)
        P_array = np.append(P_array, 0.1)
        fl_wtper = data["FluidProportion_wt"]

        while fl_wtper <= init_vapor:
            with redirect_stdout(_f):
                output = melts.equilibrate_tp(temperature, SatP_MPa, initialize=True)
            (status, temperature, p, xmlout) = output[0]
            fl_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")
            liq_mass = melts.get_mass_of_phase(xmlout, phase_name="Liquid")
            fl_comp = melts.get_composition_of_phase(xmlout, phase_name="Fluid")
            fl_wtper = 100 * fl_mass / (fl_mass + liq_mass)
            try:
                _sample_dict["H2O"] += fl_comp["H2O"] * 0.0005
            except Exception:
                _sample_dict["H2O"] = _sample_dict["H2O"] * 1.1
            try:
                _sample_dict["CO2"] += fl_comp["CO2"] * 0.0005
            except Exception:
                _sample_dict["CO2"] = _sample_dict["CO2"] * 1.1
            _sample = sample_class.Sample(_sample_dict)
            _sample_dict = _sample.get_composition(normalization="standard", units="wtpt_oxides")
            melts.set_bulk_composition(_sample_dict)  # reset MELTS

        pressure = []
        H2Oliq = []
        CO2liq = []
        H2Ofl = []
        CO2fl = []
        fluid_wtper = []
        iterno = 0
        sys.stdout.write("\r")  # carriage return to remove previous printed text
        for i in P_array:
            # Handle status_bar
            iterno += 1
            percent = iterno / len(P_array)
            batchfile.status_bar.status_bar(percent, btext="Calculating degassing path...")

            fl_mass = 0.0
            melts.set_bulk_composition(_sample_dict)
            with redirect_stdout(_f):
                output = melts.equilibrate_tp(temperature, i, initialize=True)
            (status, temperature, p, xmlout) = output[0]
            liq_comp = melts.get_composition_of_phase(xmlout, phase_name="Liquid")
            fl_comp = melts.get_composition_of_phase(xmlout, phase_name="Fluid", mode="component")
            liq_mass = melts.get_mass_of_phase(xmlout, phase_name="Liquid")
            fl_mass = melts.get_mass_of_phase(xmlout, phase_name="Fluid")
            fl_wtper = 100 * fl_mass / (fl_mass + liq_mass)

            if fl_mass > 0:
                pressure.append(p * 10.0)
                try:
                    H2Oliq.append(liq_comp["H2O"])
                except Exception:
                    H2Oliq.append(0)
                try:
                    CO2liq.append(liq_comp["CO2"])
                except Exception:
                    CO2liq.append(0)
                try:
                    H2Ofl.append(fl_comp["Water"])
                except Exception:
                    H2Ofl.append(0)
                try:
                    CO2fl.append(fl_comp["Carbon Dioxide"])
                except Exception:
                    CO2fl.append(0)
                fluid_wtper.append(fl_wtper)

                try:
                    _sample_dict["H2O"] = (liq_comp["H2O"] +
                                           (_sample_dict["H2O"] - liq_comp["H2O"]) *
                                           (1.0 - fractionate_vapor))
                except Exception:
                    _sample_dict["H2O"] = 0
                try:
                    _sample_dict["CO2"] = (liq_comp["CO2"] +
                                           (_sample_dict["CO2"] - liq_comp["CO2"]) *
                                           (1.0 - fractionate_vapor))
                except Exception:
                    _sample_dict["CO2"] = 0
            _sample = sample_class.Sample(_sample_dict)
            _sample_dict = _sample.get_composition(normalization="standard", units="wtpt_oxides")

        melts.set_bulk_composition(self.bulk_comp_orig)  # this needs to be reset always!
        open_degassing_df = pd.DataFrame(list(zip(pressure, H2Oliq, CO2liq, H2Ofl, CO2fl,
                                                  fluid_wtper)),
                                         columns=["Pressure_bars",
                                                  "H2O_liq",
                                                  "CO2_liq",
                                                  "XH2O_fl",
                                                  "XCO2_fl",
                                                  "FluidProportion_wt"])

        open_degassing_df = open_degassing_df[open_degassing_df.CO2_liq >= 0.0]
        open_degassing_df = open_degassing_df[open_degassing_df.H2O_liq >= 0.0]

        return open_degassing_df
